#------------------------------------------------------------------
# $Copyright: Copyright (c) 2017 Veritas Technologies LLC.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# VERITAS TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
#
# The  Licensed  Software  and  Documentation  are  deemed  to be commercial
# computer  software  as  defined  in  FAR  12.212 and subject to restricted
# rights  as  defined in FAR Section 52.227-19 "Commercial Computer Software
# - Restricted  Rights"  and  DFARS 227.7202,  et seq.  "Commercial Computer
# Software  and  Commercial Computer Software Documentation," as applicable,
# and any successor regulations, whether delivered by Veritas as on premises
# or  hosted  services.  Any  use,  modification,  reproduction  release,
# performance,  display  or  disclosure  of  the  Licensed  Software  and
# Documentation by the U.S. Government shall be solely in accordance with
# the terms of this Agreement. $
#
# $Header: /project/isv-agent/VProAgents/MySQL/MySQL.pm,v 1.11 2015/02/19 07:27:11 mrampura Exp $
#
# $sAuthor: $
#
# $Date: 2015/02/19 07:27:11 $
# $Revision: 1.11 $
#
# Provides:                            MySQL
# Author:                              Anshuman Atre
# crAgent Build:                       1.90 2009/06/03 07:07:19
# crAgent Run Date:                    20 Nov, 2009
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package MySQL;

use ag_i18n_inc;

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use SystemD;        # Support generalized SystemD Functions
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use Socket qw(:addrinfo);
use File::Path qw ( mkpath );
use File::stat;
use POSIX qw ( uname );

my ( $CLUS_HOME, $sPrefix, $bGlobal ) = ();

#------------------------
# Setup VCS/VCSone home..
#------------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.x/VCSOne..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvcsone";
}
else { # VCS 4.x..
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /.*(vad|vcs(|one))$/;

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdHares  = $HARES  ||= "$CLUS_HOME/bin/hares";
my $sCmdHatype = $HATYPE ||= "$CLUS_HOME/bin/hatype";
my $sCmdVcsDecrypt = $HADECRYPT ||= "$CLUS_HOME/sbin/${sPrefix}decrypt";
my @liRunningProcs = (); # Global..
my $GREP = "/bin/grep";

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: MySQL.pm,v 1.11 2015/02/19 07:27:11 mrampura Exp $::ETVHeader";
Version::TellVersion ( $sVer );

#--------------------------------
# Define MySQL subroutines..
#--------------------------------

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String EntryPointName,
                              IN String MySQLUser,
                              IN String MySQLAdmin,
                              IN String MySQLAdminPasswd,
                              IN String EnvFile,
                              IN String BaseDir,
                              IN String DataDir,
                              IN String MyCnf,
                              IN String HostName,
                              IN Integer Port,
                              IN String ServiceName,
                              IN String MySQLPidFile,
                              IN Integer SecondLevelMonitor,
                              IN String MonitorProgram );

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName               Entry pt name
 $localMySQLUser                The dedicated OS login used to run the database
 $localMySQLAdmin               The dedicated administator for the database server
 $localMySQLAdminPasswd         The encrypted password for the MySQL administrator
 $localEnvFile                  Absolute path to environment file to source
 $localBaseDir                  Absolute path to installation directory
 $localDataDir                  Absolute path to database directory
 $localMyCnf                    Absolute path to MySQL configuration file
 $localHostName                 Hostname/Bind address of virtual host for MySQL
 $localPort                     The port on which the database server listens
 $localServiceName              Name of the service to start, stop and check status of MySQL.
 $localMySQLPidFile             Full path of Pid file as specified in service file
 $localSecondLevelMonitor       Cycle count to run in-depth monitoring
 $localMonitorProgram           Full path to external monitor program
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $localMySQLUser,
       $localMySQLAdmin,
       $localMySQLAdminPasswd,
       $localEnvFile,
       $localBaseDir,
       $localDataDir,
       $localMyCnf,
       $localHostName,
       $localPort,
       $localServiceName,
       $localMySQLPidFile,
       $localSecondLevelMonitor,
       $localMonitorProgram ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, "Subroutine <%s> called with:
  EntryPointName           [%s]
  MySQLUser                [%s]
  MySQLAdmin               [%s]
  MySQLAdminPasswd         [%s]
  EnvFile                  [%s]
  BaseDir                  [%s]
  DataDir                  [%s]
  MyCnf                    [%s]
  HostName                 [%s]
  Port                     [%s]
  ServiceName              [%s]
  MySQLPidFile             [%s]
  SecondLevelMonitor       [%s]
  MonitorProgram           [%s]",

  $sMyName,  $sEntryPointName,  $localMySQLUser,  $localMySQLAdmin,
  $localMySQLAdminPasswd,       $localEnvFile,       $localBaseDir,
  $localDataDir,    $localMyCnf,    $localHostName,     $localPort,
  $localServiceName, $localMySQLPidFile, $localSecondLevelMonitor, $localMonitorProgram );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName  MySQLUser  MySQLAdmin  MySQLAdminPasswd   EnvFile
       BaseDir    DataDir   MyCnf   HostName   Port   ServiceName  
       MySQLPidFile   SecondLevelMonitor   MonitorProgram );

  @hsArgList { @lsArgs } = @_; # Assign..

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------
  my @lsOnlineArgs  = qw ( BaseDir DataDir MySQLUser
                           MySQLAdmin MySQLAdminPasswd );
  my @lsOfflineArgs = qw ( BaseDir DataDir MySQLUser );
  my @lsMonitorArgs = qw ( BaseDir DataDir MySQLUser );
  my @lsCleanArgs   = qw ( BaseDir DataDir MySQLUser );
  my @lsImfRegisterArgs = qw ( BaseDir DataDir MySQLUser );

  my %hlsNeeded = ( 'online'  => \@lsOnlineArgs,
                    'offline' => \@lsOfflineArgs,
                    'monitor' => \@lsMonitorArgs,
                    'clean'   => \@lsCleanArgs,
                    'imf_register'	=> \@lsImfRegisterArgs );

  my @lsReqdArgs = @{ $hlsNeeded { $sEntryPointName } };

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20002,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-----------------------------
  # Remove SLM file, if exists..
  #-----------------------------
  if ( $sEntryPointName ne 'monitor' ) {
    VCSagentFW::RunSecondLevelMonitor ( 1 ); # Delete..
  }

  #-------------------------------
  # Individual validation starts..
  #-------------------------------
  #-------------------------------
  # Validate parameter MySQLUser..
  #-------------------------------
  if ( ! defined getpwnam $localMySQLUser ) {
    &logError ( 20003,
    "<%s> MySQLUser [%s] is not a valid UNIX login",
    $sMyName, $localMySQLUser );
    $iReturnCode = 0; # Failure..
  }
  

  
  #-----------------------------------------------
  # Validation for online offline entry point(s)..
  #-----------------------------------------------
  if ( $sEntryPointName =~ /^(online|offline)$/o ) {
    #-----------------------------
    # Validate parameter BaseDir..
    #-----------------------------
    if ( ! -d $localBaseDir ) {
      &logError ( 20004,
      "<%s> BaseDir [%s] directory does not exist",
      $sMyName, $localBaseDir );
      $iReturnCode = 0; # Failure..
    }

    #-----------------------------
    # Validate parameter EnvFile..
    #-----------------------------
    if ( $localEnvFile =~ /\S/ && ! -e $localEnvFile ) {
      &logError ( 20005,
      "<%s> EnvFile [%s] file does not exist",
      $sMyName, $localEnvFile );
      $iReturnCode = 0; # Failure..
    }
  }

  #---------------------------------------
  # Validation for online entry point(s)..
  #---------------------------------------
  if ( $sEntryPointName eq "online" ) {
    #---------------------------
    # Validate parameter MyCnf..
    #---------------------------
    if ( $localMyCnf =~ /\S/ && ! -e $localMyCnf ) {
      &logError ( 20006,
      "<%s> MyCnf [%s] file does not exist",
      $sMyName, $localMyCnf );
      $iReturnCode = 0; # Failure..
    }

    #-----------------------------
    # Validate parameter DataDir..
    #-----------------------------
    if ( ! -d $localDataDir ) {
      &logError ( 20007,
      "<%s> DataDir [%s] directory does not exist",
      $sMyName, $localDataDir );
      $iReturnCode = 0; # Failure..
    }

    #---------------------------------------
    # Validate parameter HostName and Port..
    #---------------------------------------
    
    if ( $localSecondLevelMonitor ) {
      if ( $localHostName !~ /\S/ ) {
        &logError ( 20008,
        "<%s> HostName [%s] cannot be NULL, if SLM is enabled",
        $sMyName, $localHostName );
        $iReturnCode = 0; # Failure..
      }
      else {
        my ($addr_err, @res) = getaddrinfo($localHostName);
        if ( ! @res ) { 
          &logError ( 20009,
          "<%s> HostName [%s] is an invalid Hostname/IPv4/IPv6 address",
          $sMyName, $localHostName );
          $iReturnCode = 0; # Failure..
          }
      }
  
      #--------------------------
      # Validate parameter Port..
      #--------------------------
      if ( ! $localPort ) {
        &logError ( 20010,
        "<%s> Port [%s] has to be greater than 0, if SLM is enabled",
        $sMyName, $localPort );
        $iReturnCode = 0; # Failure..
      }
    }

    #------------------------------------
    # Validate parameter MonitorProgram..
    #------------------------------------
    if ( $localMonitorProgram =~ /(\S+)/ && # Defined..
         ! -x $1 ) {
      &logError ( 20011,
      "<%s> MonitorProgram [%s] is not executable",
      $sMyName, $1 );
      $iReturnCode = 0; # Failure..
    }
  }

  $iReturnCode ||
  &logWarn ( 20012,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN String BaseDir,
                         IN String DataDir,
                         IN String EnvFile,
                         IN String HostName,
                         IN String MyCnf,
                         IN String MySQLAdmin,
                         IN String MySQLAdminPasswd,
                         IN String MySQLUser,
                         IN Integer Port,
                         IN String ServiceName,
                         IN String MySQLPidFile,
                         IN Integer SecondLevelMonitor );

 Description : Start logic needed for online entry pt

 Parameters :

 $sBaseDir                  Absolute path to installation directory
 $sDataDir                  Absolute path to database directory
 $sEnvFile                  Absolute path to environment file to source
 $sHostName                 Hostname/Bind address of virtual host for MySQL
 $sMyCnf                    Absolute path to MySQL configuration file
 $sMySQLAdmin               The dedicated administator for the database server
 $sMySQLAdminPasswd         The encrypted password for the MySQL administrator
 $sMySQLUser                The dedicated OS login used to run the database
 $iPort                     The port on which the database server listens
 $ServiceName              Name of the service to start, stop and check status of MySQL.
 $MySQLPidFile             Full path of Pid file as specified in service file
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sBaseDir,
       $sDataDir,
       $sEnvFile,
       $sHostName,
       $sMyCnf,
       $sMySQLAdmin,
       $sMySQLAdminPasswd,
       $sMySQLUser,
       $iPort,
       $sServiceName,
       $sMySQLPidFile,
       $iSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20013, "Subroutine <%s> called with:
  BaseDir                  [%s]
  DataDir                  [%s]
  EnvFile                  [%s]
  HostName                 [%s]
  MyCnf                    [%s]
  MySQLAdmin               [%s]
  MySQLAdminPasswd         [%s]
  MySQLUser                [%s]
  Port                     [%s]
  ServiceName              [%s]
  MySQLPidFile             [%s]
  SecondLevelMonitor       [%s]",

  $sMyName, $sBaseDir, $sDataDir, $sEnvFile,  $sHostName,  $sMyCnf,
  $sMySQLAdmin,     $sMySQLAdminPasswd,     $sMySQLUser,    $iPort,
  $sServiceName, $sMySQLPidFile, $iSecondLevelMonitor );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,
       @lsCmdout, @liPids ) = ();

  #-----------------------------------------
  # Whack any preexisting MySQL processes..
  #-----------------------------------------
  &logTrace ( 20014,
  "<%s> Killing MySQL processes for this instance", $sMyName );

  if ( MySQL::KillMySQLProcesses ( $sBaseDir,
                                   $sDataDir,
                                   $sHostName,
                                   $sMySQLUser,
                                   $sServiceName,
                                   $sMySQLPidFile ) ) {
    &logTrace ( 20015,
    "<%s> Processes for this instance do not exist before start-up",
    $sMyName );
  }
  else {
    &logError ( 20016,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }
  
  #--------------------
  # Return, if failed..
  #--------------------
  $iReturnCode || return $iReturnCode;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + MySQL::GetEngineTimeout ( 1 );
  my $iRet = ();
  $iRet = SystemD::DetectSystemd();
  
  if (! $iRet ) {
    #------------------------------------------
    # Define start command and args variables..
    #------------------------------------------
    $sCommand = "$sBaseDir/bin/mysqld_safe";
  
    #--------------------------------------
    # Use optional parameters, if defined..
    #--------------------------------------
    $sCommandArgs = "--defaults-file=$sMyCnf " if ( $sMyCnf =~ /\S/ );
    $sCommandArgs .= "--datadir=$sDataDir --user=$sMySQLUser &";
    ( @lsCmdout, $iExitCode, $iTimeout ) = ();
  
    &logTrace ( 20017,
    "<%s> Starting MySQL instance with Cmd [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  }
  
  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20018,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iRet ) {
    my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
    if ( $sServiceName !~ /\S/ ) {
      $sServiceName = "mysqld";
      if ( -e "/etc/SuSE-release" ) {
        $sServiceName = "mysql";
      }
    }
    if( $iReturnCode = SystemD::RunSystemdServiceName($sResourceName, $sServiceName, "start" )){
        &logTrace ( 20323,
        "<%s> RunSystemd was successful",
        $sMyName );
      }
  }
  elsif ( $iReturnCode =
    MySQL::ExecLocalCmd ( $sMySQLUser, 
                          $sEnvFile, 
                          $sCommand, 
                          $sCommandArgs, 
                          \$iExitCode, 
                          $iTimeout, 
                          \@lsCmdout,
				"",
                          $sBaseDir ) ) {
    &logTrace ( 20019,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;

  #--------------------
  # Delay entry point..
  #--------------------
  my $iSleepTime = 10; # Secs..
  
  while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
    @liPids = (); # Initialise..

    sleep $iSleepTime; # Sec..

    &logTrace ( 20020,
    "<%s> Confirming start-up with [%s] sec in hand",
    $sMyName, $iTimeout - $iSleepTime );

    #---------
    # Do FLM..
    #---------
    if ( VCSDef_MonitorExit::ONLINE ==
         MySQL::GetProcState ( \@liPids,
                               $sBaseDir,
                               $sDataDir,
                               $sHostName,
                               $sMySQLUser,
                               $sServiceName,
                               $sMySQLPidFile,
                               1 ) && (         # Running FFLM..
         $iSecondLevelMonitor  ?                # If enabled..
         VCSDef_MonitorExit::ONLINE ==          # Online after..
         MySQL::SecondLevelMonitor ( $sBaseDir, # Running SLM..
                                     $sEnvFile,
                                     $sHostName,
                                     $sMyCnf,
                                     $sMySQLAdmin,
                                     $sMySQLAdminPasswd,
                                     $sMySQLUser,
                                     $iPort,
                                     $sServiceName,
                                     $iTillWhen ) : 1 ) ) { 
      #---------------------
      # Start-up confirmed..
      #---------------------
      &logInfo ( 20021,
      "<%s> Confirmed start-up of this instance of MySQL server",
      $sMyName );

      $iReturnCode = 1; # Success..
      last;
    }
    else {
      &logTrace ( 20022,
      "<%s> Could not confirm start-up. Sleeping for [%s] sec",
      $sMyName, $iSleepTime );

      $iReturnCode = 0; # Failure..
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN String BaseDir,
                        IN String DataDir,
                        IN String EnvFile,
                        IN String HostName,
                        IN String MyCnf,
                        IN String MySQLAdmin,
                        IN String MySQLAdminPasswd,
                        IN String MySQLUser,
                        IN String ServiceName,
                        IN String MySQLPidFile );

 Description : Stop logic needed for offline entry pt

 Parameters :

 $sBaseDir                  Absolute path to installation directory
 $sDataDir                  Absolute path to database directory
 $sEnvFile                  Absolute path to environment file to source
 $sHostName                 Hostname/Bind address of virtual host for MySQL
 $sMyCnf                    Absolute path to MySQL configuration file
 $sMySQLAdmin               The dedicated administator for the database server
 $sMySQLAdminPasswd         The encrypted password for the MySQL administrator
 $sMySQLUser                The dedicated OS login used to run the database
 $ServiceName              Name of the service to start, stop and check status of MySQL. 
 $MySQLPidFile             Full path of Pid file as specified in service file
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sBaseDir,
       $sDataDir,
       $sEnvFile,
       $sHostName,
       $sMyCnf,
       $sMySQLAdmin,
       $sMySQLAdminPasswd,
       $sMySQLUser,
       $sServiceName,
       $sMySQLPidFile ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20023, "Subroutine <%s> called with:
  BaseDir                  [%s]
  DataDir                  [%s]
  EnvFile                  [%s]
  HostName                 [%s]
  MyCnf                    [%s]
  MySQLAdmin               [%s]
  MySQLAdminPasswd         [%s]
  MySQLUser                [%s]
  ServiceName              [%s]
  MySQLPidFile             [%s]",

  $sMyName,   $sBaseDir,    $sDataDir,    $sEnvFile,    $sHostName,
  $sMyCnf, $sMySQLAdmin, $sMySQLAdminPasswd, $sMySQLUser, $sServiceName, $sMySQLPidFile );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + MySQL::GetEngineTimeout ( 1 ); # With margin..

  #------------------------------------------
  # Declare stop command and args variables..
  #------------------------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode, $iSleepTime,
       @lsCmdout, @liPids ) = ();
  my $iRet = ();
  $iRet = SystemD::DetectSystemd();

  if (! $iRet ) {
    #------------------------------------
    # Define stop command and arguments..
    #------------------------------------
    $sCommand = "$sBaseDir/bin/mysqladmin";
  
    #--------------------------------------
    # Use optional parameters, if defined..
    #--------------------------------------
    $sCommandArgs = "--defaults-file=$sMyCnf " if ( $sMyCnf =~ /\S/ );
    $sCommandArgs .= "--user=$sMySQLAdmin " .
                     "--password=$sMySQLAdminPasswd shutdown";
    ( $iExitCode, $iTimeout, @lsCmdout ) = (); # Truncate.. 
  
    &logTrace ( 20024,
    "<%s> Stopping MySQL with Command [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  }  
  #------------------------
  # Request shutdown..
  #------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20025,
    "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iRet ) {
    my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
    if ( $sServiceName !~ /\S/ ){
      $sServiceName = "mysqld";
      if ( -e "/etc/SuSE-release" ) {
        $sServiceName = "mysql";
      }
    }

    if( $iReturnCode = SystemD::RunSystemdServiceName($sResourceName, $sServiceName, "stop" )){
        &logTrace ( 20324,
        "<%s> RunSystemd was successful",
        $sMyName );
      }
  }

  elsif ( $iReturnCode =
          MySQL::ExecLocalCmd ( $sMySQLUser, 
                                $sEnvFile, 
                                $sCommand, 
                                $sCommandArgs, 
                                \$iExitCode, 
                                $iTimeout, 
                                \@lsCmdout, 
                                $sMySQLAdminPasswd ) ) {
    &logTrace ( 20026,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }

  #--------------------
  # Delay entry point..
  #--------------------
  if ( $iReturnCode ) { # Success..
    #-----------------
    # Set sleep time..
    #-----------------
    $iSleepTime = 10; # Sec..
  
    &logInfo ( 20027,
    "<%s> Delaying exit to confirm shutdown..", $sMyName );

    #--------------------
    # Delay entry point..
    #--------------------
    while ( ( $iTimeout = $iTillWhen - time ) > 20 ) {
      #-----------------------
      # Wait before retrying..
      #-----------------------
      sleep $iSleepTime; # Secs..
  
      MySQL::GetProcState ( \@liPids,
                            $sBaseDir,
                            $sDataDir,
                            $sHostName,
                            $sMySQLUser,
                            $sServiceName,
                            $sMySQLPidFile,
                            0 ); # Request regular..
  
      #------------------------------
      # Check for pending processes..
      #------------------------------
      if ( scalar @liPids ) {
        &logTrace ( 20028,
        "<%s> Could not confirm shutdown. Retrying in [%s] sec.
        Delaying kill with [%s] sec in hand",
        $sMyName, $iSleepTime, $iTimeout - $iSleepTime );
      }
      else {
        &logInfo ( 20029,
        "<%s> Confirmed shutdown of MySQL instance",
        $sMyName );
        last; # Get out..
      }
    }
  }

  #-----------------------------------------
  # Whack any persisting MySQL processes..
  #-----------------------------------------
  &logTrace ( 20030,
  "<%s> Killing MySQL processes for this instance", $sMyName );

  if ( MySQL::KillMySQLProcesses ( $sBaseDir,
                                   $sDataDir,
                                   $sHostName,
                                   $sMySQLUser,
                                   $sServiceName,
                                   $sMySQLPidFile ) ) {
    &logTrace ( 20031,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20032,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }
  
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN String BaseDir,
                        IN Integer CleanReason,
                        IN String DataDir,
                        IN String EnvFile,
                        IN String HostName,
                        IN String MyCnf,
                        IN String MySQLAdmin,
                        IN String MySQLAdminPasswd,
                        IN String MySQLUser,
                        IN String ServiceName,
                        IN String MySQLPidFile );

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $sBaseDir                  Absolute path to installation directory
 $iCleanReason              Why clean was called
 $sDataDir                  Absolute path to database directory
 $sEnvFile                  Absolute path to environment file to source
 $sHostName                 Hostname/Bind address of virtual host for MySQL
 $sMyCnf                    Absolute path to MySQL configuration file
 $sMySQLAdmin               The dedicated administator for the database server
 $sMySQLAdminPasswd         The encrypted password for the MySQL administrator
 $sMySQLUser                The dedicated OS login used to run the database
 $ServiceName              Name of the service to start, stop and check status of MySQL.
 $MySQLPidFile             Full path of Pid file as specified in service file 

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sBaseDir,
       $iCleanReason,
       $sDataDir,
       $sEnvFile,
       $sHostName,
       $sMyCnf,
       $sMySQLAdmin,
       $sMySQLAdminPasswd,
       $sMySQLUser,
       $sServiceName,
       $sMySQLPidFile ) = @_;

  #-----------------------------
  # Figure out why were called..
  #-----------------------------
  my ( $sCleanReason, %hsCleanReasons ) = ();
  my @lsCleanStrings = qw ( Offline_timed_out    Offline_was_ineffective
                            Online_timed_out      Online_was_ineffective
                            Found_offline_unexpectedly Monitor_timed_out );

  @hsCleanReasons { 0 .. $#lsCleanStrings } = @lsCleanStrings; # Populate..
  ( $sCleanReason = $hsCleanReasons { $iCleanReason } ||       # Reason..
                    $iCleanReason ) =~ s/_/ /g;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20033, "Subroutine <%s> called with:
  BaseDir                  [%s]
  CleanReason              [%s]
  DataDir                  [%s]
  EnvFile                  [%s]
  HostName                 [%s]
  MyCnf                    [%s]
  MySQLAdmin               [%s]
  MySQLAdminPasswd         [%s]
  MySQLUser                [%s]
  ServiceName              [%s]
  MySQLPidFile             [%s]",

  $sMyName,   $sBaseDir,   $sCleanReason,   $sDataDir,   $sEnvFile,
  $sHostName, $sMyCnf, $sMySQLAdmin, $sMySQLAdminPasswd, $sMySQLUser, $sServiceName, $sMySQLPidFile );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + MySQL::GetEngineTimeout ( 1 ); # With margin..

  #--------------------------
  # Declare local variables..
  #--------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode,
       @lsFiles, @lsCmdout ) = ();
  
  #-------------------------------------
  # Define clean command and arguments..
  #-------------------------------------
  $sCommand = "$sBaseDir/bin/mysqladmin";
  $sCommandArgs = "--defaults-file=$sMyCnf " if ( $sMyCnf =~ /\S/ );
  $sCommandArgs .= "--user=$sMySQLAdmin " .
                  "--password=$sMySQLAdminPasswd shutdown";

  unless ( $hsCleanReasons { $iCleanReason } eq 'Offline_timed_out' ||
           $hsCleanReasons { $iCleanReason } eq 'Offline_was_ineffective' ) {
    #-------------------
    # Try to be gentle..
    #-------------------
    my $iRet = ();
    $iRet = SystemD::DetectSystemd();
    if (! $iRet ) {
      &logInfo ( 20034,
      "<%s> Attempting to stop MySQL with Command [%s] Args [%s]",
      $sMyName, $sCommand, $sCommandArgs );
    }
    #------------------------
    # Request shutdown..
    #------------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20035,
      "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ($iRet) {
      my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
      if ( $sServiceName !~ /\S/ ) {
        $sServiceName = "mysqld";
        if ( -e "/etc/SuSE-release" ) {
          $sServiceName = "mysql";
        }
      }
      if( $iReturnCode = SystemD::RunSystemdServiceName($sResourceName, $sServiceName, "stop" )){
        &logTrace ( 20326,
          "<%s> RunSystemd was successful",
          $sMyName );
      }
    }
    elsif ( MySQL::ExecLocalCmd ( $sMySQLUser, 
                                  $sEnvFile, 
                                  $sCommand, 
                                  $sCommandArgs, 
                                  \$iExitCode, 
                                  $iTimeout, 
                                  \@lsCmdout, 
                                  $sMySQLAdminPasswd ) ) {
      &logTrace ( 20036, 
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
  }

  #-----------------------------------------
  # Whack any persisting MySQL processes..
  #-----------------------------------------
  &logTrace ( 20037,
  "<%s> Killing MySQL processes for this instance", $sMyName );

  if ( MySQL::KillMySQLProcesses ( $sBaseDir,
                                   $sDataDir,
                                   $sHostName,
                                   $sMySQLUser,
                                   $sServiceName,
                                   $sMySQLPidFile ) ) {
    &logTrace ( 20038,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20039,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }
  
  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( IN String BaseDir,
                                   IN String DataDir,
                                   IN String HostName,
                                   IN String MySQLUser,
                                   IN String ServiceName,
                                   IN String MySQLPidFile );

 Description : First level monitor to determine state of resource.

 Parameters :

 $sBaseDir                  Absolute path to installation directory
 $sDataDir                  Absolute path to database directory
 $sHostName                 Hostname/Bind address of virtual host for MySQL
 $sMySQLUser                The dedicated OS login used to run the database
 $ServiceName               Name of the service to start, stop and check status of MySQL.
 $MySQLPidFile              Full path of Pid file as specified in service file 

 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sBaseDir,
       $sDataDir,
       $sHostName,
       $sMySQLUser,
       $sServiceName,
       $sMySQLPidFile ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20040, "Subroutine <%s> called:
  BaseDir                  [%s]
  DataDir                  [%s]
  HostName                 [%s]
  MySQLUser                [%s]
  ServiceName              [%s]
  MySQLPidFile             [%s]",
  $sMyName, $sBaseDir, $sDataDir, $sHostName, $sMySQLUser, $sServiceName, $sMySQLPidFile );
  
  my @liDummyPids = ();
  my $iReturnCode = MySQL::GetProcState ( \@liDummyPids,
                                          $sBaseDir,
                                          $sDataDir,
                                          $sHostName,
                                          $sMySQLUser,
                                          $sServiceName,
                                          $sMySQLPidFile,
                                          1 ); # Use FFLM..

  if ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20041,
    "<%s> Found all required processes. Return code set to [%s]",
    $sMyName, $iReturnCode );
  }
  else {
    &logTrace ( 20042,
    "<%s> Did not find all required processes. Returning [%s]",
    $sMyName, $iReturnCode );
  }

  #--------------------------------
  # Populate the global PID array..
  #--------------------------------
  @liRunningProcs = @liDummyPids;
                                                                                
  &logTrace ( 20043,
  "<%s> Populating global running proclist with [%s] PIDs [%s]",
  $sMyName, scalar @liRunningProcs, join ( ' ', @liRunningProcs ) );

  return ( $iReturnCode );
}


=pod

 -------------------------------------------------------------------------
 Function : int GetProcState ( OUT ListReference OnlinePids,
                               IN String BaseDir,
                               IN String DataDir,
                               IN String HostName,
                               IN String ServiceName,
                               IN String MySQLUser,
                               IN String MySQLPidFile,
                               IN Integer UseFFLM );

 Description : First level monitor to determine state of resource.

 Parameters :

 $rliOnlinePids             PIDs of relevant running processes
 $sBaseDir                  Absolute path to installation directory
 $sDataDir                  Absolute path to database directory
 $sHostName                 Hostname/Bind address of virtual host for MySQL
 $sMySQLUser                The dedicated OS login used to run the database
 $ServiceName               Name of the service to start, stop and check status of MySQL.
 $MySQLPidFile              Full path of Pid file as specified in service file
 $iUseFFLM                  Filtering procedure
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetProcState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rliOnlinePids,
       $sBaseDir,
       $sDataDir,
       $sHostName,
       $sMySQLUser,
       $sServiceName,
       $sMySQLPidFile,
       $iUseFFLM ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20044, "Subroutine <%s> called with:
  OnlinePids               [%s]
  BaseDir                  [%s]
  DataDir                  [%s]
  HostName                 [%s]
  MySQLUser                [%s]
  ServiceName              [%s]
  MySQLPidFile             [%s]
  UseFFLM                  [%s]",

  $sMyName,  $rliOnlinePids,  $sBaseDir,   $sDataDir,   $sHostName,
  $sMySQLUser, $sServiceName, $sMySQLPidFile, $iUseFFLM );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #----------------------
  # Initialise Variables..
  #----------------------
  my ( $iProcessCount, $iUseRFLM, $iFound, $iUid, $iPid, $sLang,
       $sKey, $rhlProc, $sPidFile, $sProc, $sTemp, $sOSVer,
       $sFile, $sPointsTo, @hProcListHash, @lsUpdates, @lsValues,
       %hiProcPids, %hiPids, %hsSrvrDetails, %hsCopy, %hiLocal ) = ();

  my %hsZoneVers = map { $_ => 1 } qw ( 5.10 5.11 ); # Keep adding here..
  @$rliOnlinePids = (); # Truncate..

  #--------------------------
  # Other processes to kill..
  #--------------------------
  my @lsOthers = ( qr#mysqld(.*--pid-file=$sMySQLPidFile\b|.*--datadir=$sDataDir\b|$)# );
  if (SystemD::DetectSystemd()) {
    if ( $sServiceName =~ /@(\S+)\b/ ) {
      my $sSuffix = $1;
      @lsOthers = ( qr#mysql.*--defaults-group-suffix=\@$sSuffix\b# );
    }
  }
  #--------------------------------------------------
  # Attempt to get environment of running processes..
  #--------------------------------------------------
  if ( $^O eq 'solaris' ) {
    $sOSVer = ( uname ) [2]; # Get OS version..

    #--------------------------------
    # Handle Solaris 10 local zones..
    #--------------------------------
    if ( $hsZoneVers { $sOSVer } ) { # Solaris 10 or 11..
      if ( kill ( 0, 1 ) ) {         # Inside global zone..
        ++ $bGlobal;                 # Set global..
        &logTrace ( 20045,
        "<%s> Running inside OS [%s] OSVer [%s]. Global flag set to [%s]",
        $sMyName, $^O, $sOSVer, $bGlobal );
      }
      else { # Inside local zone..
        $bGlobal = 0;
        &logTrace ( 20046,
        "<%s> Running inside OS [%s] OSVer [%s] inside local zone",
        $sMyName, $^O, $sOSVer );
      }
    }
  }

  #------------------------------------
  # Attempt to get monitoring details..
  #------------------------------------
  if ( ( 0 == ( $iUid = getpwnam $sMySQLUser ) ) && # User is the super-user..
       ( 2 == $iUseFFLM ) ) {                       # And called to kill-by-UID..
    &logTrace ( 20047, "<%s> Unsafe to kill processes by User [%s] UID [%s].
    Using filter-based approach.", $sMyName, $sMySQLUser, $iUid ),
    $iUseFFLM = 1;
  }

  if ( $iUseFFLM != 2 ) { # Called for first level check..
    if ( MySQL::GetSrvrDetails ( $sBaseDir, 
                                 $sDataDir, 
                                 $sHostName,
                                 $sServiceName,
                                 $sMySQLPidFile, 
                                 \%hsSrvrDetails ) ) {
      &logTrace ( 20048,
      "<%s> Got details for monitoring MySQL instance", $sMyName );
  
      %hsCopy = %hsSrvrDetails;        # Backup..
      $iUseRFLM = scalar keys %hsCopy; # Set to entire count..
    }
    else {
      &logError ( 20049,
      "<%s> Failed to get details for monitoring MySQL instance",
      $sMyName );
      return VCSDef_MonitorExit::UNKNOWN;
    }
                                  
    #---------------------------------
    # Check if we attempt to be fast.. 
    #---------------------------------
    if ( $iUseFFLM ) { # FFLM..
      $iUseRFLM = 0;   # Hopefully skip..

      foreach $sKey ( keys %hsCopy ) {          # Use copy..
        @lsUpdates = ();                        # Reset..
        @lsValues = @{ $hsCopy { $sKey } };     # Get values..
        $sTemp .= "\nMySQL process [$sKey]\n";  # Display string..

        #----------------
        # Read in pairs..
        #----------------
        while ( ( $sPidFile, $sProc ) = splice ( @lsValues, 0, 2 ) ) {

          #-----------
          # Run FFLM..
          #-----------
          $iFound = MySQL::FFLMStatus ( $sPidFile,
                                        $sProc,
                                        $iUid,
                                        \$iPid );

          #----------------
          # Update status..
          #----------------
          push ( @lsUpdates, $sPidFile, $sProc, $iFound );

          $sTemp .= "PID File     [$sPidFile]\n";
          $sTemp .= "Srch Process [$sProc]\n";
          $sTemp .= "FFLM Status  [$iFound]\n\n";
        }

        @{ $hsCopy { $sKey } } = @lsUpdates;

        #------------------------
        # Handle monitor status..
        #------------------------
        if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
          push ( @$rliOnlinePids, $iPid ); # Push..
          ++ $hiPids { $iPid }; # Store..
        }
        else {
          ++ $iUseRFLM; # Mark it for RFLM..
        }
      }

      #----------------
      # Print summary..
      #----------------
      &logTrace ( 20050,
      "<%s> MySQL instance process details ---%s\n\n",
      $sMyName, $sTemp );

      ( $sTemp, $sPidFile, $sProc ) = (); # Release memory..
    }
    else { # Use RLFM..
      #-----------------------------------------
      # Mark status of each service as offline..
      #-----------------------------------------
      $iFound = VCSDef_MonitorExit::OFFLINE; # Set to offline..
      foreach $sKey ( keys %hsCopy ) {
        @lsUpdates = (); # Reset..

        @lsValues = @{ $hsCopy { $sKey } }; # Get values..
        while ( scalar @lsValues ) {
          push ( @lsUpdates, splice ( @lsValues, 0, 2 ), $iFound );
        }

        @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
      }
    }
  }
  else { # Called for Kill..
    &logTrace ( 20051,
    "<%s> Called to get processes for User [%s] UID [%s]",
    $sMyName, $sMySQLUser, $iUid );
    ++ $iUseRFLM;
  }

  #-----------------------
  # Regular slow :-( FLM..
  #-----------------------
  if ( $iUseRFLM ) {
    &logTrace ( 20052,
    "<%s> Running regular first level monitor", $sMyName );

    #----------------------------------
    # Set LANG for better performance..
    #----------------------------------
    $sLang = $ENV { LANG }; # Backup..
    $ENV { LANG } = 'C';    # Now set..

    #-----------------------------------
    # Call processes and store in hash..
    #-----------------------------------
    $iProcessCount = Proc::GetLongProcessListHash ( \@hProcListHash );

    #----------
    # Restore..
    #----------
    defined $sLang ? $ENV { LANG } = $sLang : delete $ENV { LANG }; # Restore..

    #----------------------
    # Check if we got any..
    #----------------------
    if ( $iProcessCount ) {
      &logTrace ( 20053,
      "<%s> Got [%s] processes from proc table for filtering",
      $sMyName, $iProcessCount );
    }
    else {
      &logError ( 20054,
      "<%s> Call to get process list did not return any processes",
      $sMyName );
      return VCSDef_MonitorExit::OFFLINE; # Offline..
    }

    #-----------------------
    # Search for processes..
    #-----------------------
    foreach $rhlProc ( @hProcListHash ) {
      #--------------------------------------------------------------
      # Each $rhlProc is a ref. to a hash with the following keys..
      # "UID"     =>  <user id>
      # "PID"     =>  <process id>
      # "PPID"    =>  <parent process id>
      # "COMMAND" =>  <command string>
      #--------------------------------------------------------------

      next unless defined $$rhlProc { UID }; # Bug in Proc::GetLongProcessListHash..
      next if $$rhlProc { PID } == $$;       # Skip self..
      $iPid = $$rhlProc { PID };             # Assign..

      &logTrace ( 20055, "<%s> Searching UID [%s] PID [%s] PPID [%s] CMD [%s]",
      $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
      $$rhlProc { COMMAND } );

      if ( $$rhlProc { UID } == $iUid ) { # Process owner..
        #-----------------------------------------------
        # Look for processes that were offline in FFLM..
        #-----------------------------------------------
        if ( $iUseFFLM != 2 ) { # Request to filter..
          foreach $sKey ( keys %hsCopy ) {
            @lsUpdates = (); # Reset..
            @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    
            #--------------------------------------------------
            # Should have details per three elements of array..
            # <PIDFILE> <PROCESS> <STATE>
            #--------------------------------------------------
            while ( ( $sPidFile, $sProc, $iFound ) =
                      splice ( @lsValues, 0, 3 ) ) {
              if ( VCSDef_MonitorExit::OFFLINE == $iFound ) { # Earlier marked offline..
                if ( $$rhlProc { COMMAND } =~ /$sProc/ ) {    # Now found..

                  #-----------------------
                  # Handle Solaris zones..
                  #-----------------------
                  if ( $bGlobal ) { # In global zone..
                    $sFile = "/proc/$iPid/path/root";
                    if ( ( $sPointsTo = readlink $sFile ) && $sPointsTo ne "/" ) {
                      $sPointsTo = ( $sPointsTo =~ m@[^/]+@g ) [-2]; # Get zone name..
                      &logTrace ( 20056,
                      "<%s> Skipping PID [%s] PPID [%s] CMD [%s] of zone [%s]",
                      $sMyName, $iPid, $$rhlProc { PPID },  $$rhlProc { COMMAND },
                      $sPointsTo );
                      ++ $hiLocal { $iPid }; # Store..
                      last; # Get next process..
                    }
                  }

                  unshift ( @$rliOnlinePids, $iPid );         # Store..
    
                  &logTrace ( 20057,
                  "<%s> Found PID [%s] PPID [%s] CMD [%s] with filter [%s]",
                  $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND },
                  $sProc );

                  ++ $hiPids { $iPid };                 # Store..
                  $hiProcPids { $sKey } = $iPid;        # Store..
                  $iFound = VCSDef_MonitorExit::ONLINE; # Mark online..
                }
              }
    
              #----------------
              # Update status..
              #----------------
              push ( @lsUpdates, $sPidFile, $sProc, $iFound );
            }
    
            @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
          }
        }
        else { # Request for kill..
          &logTrace ( 20058,
          "<%s> Pushing PID [%s] with PPID [%s] Cmd [%s] on stack",
          $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
          push ( @$rliOnlinePids, $iPid ); # Store..
        } 
      }
      
      #---------------------------------
      # Handle other related processes..
      #---------------------------------
      if ( ( ! $hiPids  { $iPid } )     &&                     # Not caught earlier..
           ( ! $hiLocal { $iPid } )     &&                     # And not of a local zone.. 
           ( $$rhlProc { UID } == $iUid ||                     # Owned by MySQLUser
             0 == $$rhlProc { UID } )   &&                     # or owned by root..
           map { $$rhlProc { COMMAND } =~ /$_/ } @lsOthers ) { # Was now found..
        push ( @$rliOnlinePids, $iPid );                       # Store..

        &logTrace ( 20059,
        "<%s> Matched process PID [%s] PPID [%s] CMD [%s]",
        $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
      }
    }
  }

  #------------------------------------
  # Set final status for each process..
  #------------------------------------
  my ( $iOnline, $iOffline, $sEntryPointName ) = ();

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..
  $sTemp = ''; # Ensure..

  foreach $sKey ( keys %hsCopy ) {
    @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    $sTemp .= "\nProcess      [$sKey]\n";

    while ( ( $sPidFile, $sProc, $iFound ) =
              splice ( @lsValues, 0, 3 ) ) {
      #-----------------------------------
      # Create PID files for use in FFLM..
      #-----------------------------------
      if ( $iUseRFLM                                 &&      # In RFLM..
           $iUseFFLM != 2                            &&      # Asked to filter..
           $hiProcPids { $sKey }                     &&      # PID populated..
           ( $sEntryPointName =~ /^(monitor|online)$/o ) ) { # In monitor or online..

        &logTrace ( 20060,
        "<%s> Creating PID file [%s] with PID [%s]",
        $sMyName, $sPidFile, $hiProcPids { $sKey } );

        #---------
        # Create..
        #---------
        if ( MySQL::CreatePidFile ( $sPidFile,
                                     $hiProcPids { $sKey } ) ) {
          &logTrace ( 20061,
          "<%s> Successfully created PID file [%s]", $sMyName, $sPidFile );
        }
        else {
          &logError ( 20062,
          "<%s> Could not create PID file [%s] successfully",
          $sMyName, $sPidFile );
        }
      }

      #-------------
      # Set status..
      #-------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
        $iOnline = $iFound;
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) {
        $iOffline = $iFound;
      }

      $sTemp .= "Srch Process [$sProc]\n";
      $sTemp .= "FLM Status   [$iFound]\n\n";
    }

    $sTemp .= "--------------------";
  }

  #-------------------
  # Set final status..
  #-------------------
  $iReturnCode = $iOffline || $iOnline || VCSDef_MonitorExit::UNKNOWN;

  #------------------------------------------------
  # Reset status based on PIDs, if not in monitor..
  #------------------------------------------------
  if ( $sEntryPointName ne 'monitor'                   &&
       ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ) &&
       scalar @$rliOnlinePids ) {
    $iReturnCode = VCSDef_MonitorExit::PARTIAL_ONLINE;

    &logTrace ( 20063,
    "<%s> Found [%s] PIDs [%s] for this instance of MySQL.
    Resetting return code to [%s]",
    $sMyName, scalar @$rliOnlinePids, join ( ' ', @$rliOnlinePids ),
    $iReturnCode );
  }

  #-----------------------
  # Report on FLM status..
  #-----------------------
  if ( $iUseFFLM != 2 ) { # Request to filter..
    &logTrace ( 20064,
    "<%s> MySQL process details ---\n%s\nReturn code is [%s]",
    $sMyName, $sTemp, $iReturnCode );
    $sTemp = undef; # Release..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int Monitor ( IN String BaseDir,
                          IN String EnvFile,
                          IN String HostName,
                          IN String MyCnf,
                          IN String MonitorProgram,
                          IN String MySQLAdmin,
                          IN String MySQLAdminPasswd,
                          IN String MySQLUser,
                          IN Integer Port,
                          IN String ServiceName,
                          IN Integer SecondLevelMonitor );

 Description: Performs the in-depth monitoring

 Parameters :

 $sBaseDir                  Absolute path to installation directory
 $sEnvFile                  Absolute path to environment file to source
 $sHostName                 Hostname/Bind address of virtual host for MySQL
 $sMyCnf                    Absolute path to MySQL configuration file
 $sMonitorProgram           Full path to external monitor program
 $sMySQLAdmin               The dedicated administator for the database server
 $sMySQLAdminPasswd         The encrypted password for the MySQL administrator
 $sMySQLUser                The dedicated OS login used to run the database
 $iPort                     The port on which the database server listens
 $ServiceName              Name of the service to start, stop and check status of MySQL.
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if in-depth monitoring cannot be completed.
 -------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sBaseDir,
       $sEnvFile,
       $sHostName,
       $sMyCnf,
       $sMonitorProgram,
       $sMySQLAdmin,
       $sMySQLAdminPasswd,
       $sMySQLUser,
       $iPort,
       $sServiceName,
       $iSecondLevelMonitor,
       $level_two ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20065, "Subroutine <%s> called with:
  BaseDir                  [%s]
  EnvFile                  [%s]
  HostName                 [%s]
  MyCnf                    [%s]
  MonitorProgram           [%s]
  MySQLAdmin               [%s]
  MySQLAdminPasswd         [%s]
  MySQLUser                [%s]
  Port                     [%s]
  ServiceName              [%s]
  SecondLevelMonitor       [%s]
  level_two                [%s]",

  $sMyName,  $sBaseDir,  $sEnvFile,  $sHostName, $sMyCnf,  $sMonitorProgram,
  $sMySQLAdmin,     $sMySQLAdminPasswd,     $sMySQLUser,    $iPort,   $sServiceName, 
  $iSecondLevelMonitor, $level_two );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  my $iTillWhen = undef;

  #---------------------------------------
  # Do Second level monitoring if enabled.
  #---------------------------------------
  if ( -300 == $level_two ) {
     $level_two = VCSagentFW::RunSecondLevelMonitor ( $iSecondLevelMonitor );  
     &logTrace ( 20310, "<%s> Traditional second level monitoring, SLM flag for this monitor set to [%s] ", $sMyName, $level_two  );
  }
  #----------------------------------------------
  # In-depth monitoring, if correct time to run..
  #----------------------------------------------
  if ( $level_two >= 1 ) 
  {
    #--------------------------
    # Just In Time Validation..
    #--------------------------
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         MySQL::DelayedAttributeCheck ( $sBaseDir,
                                        $sEnvFile,
                                        $sHostName,
                                        $sMySQLAdmin,
                                        $sMySQLAdminPasswd,
                                        $sMySQLUser,
                                        $iPort ) ) ) {
      &logTrace ( 20066,
      "<%s> Delayed attribute check completed successfully",
      $sMyName );
    }
    else {
      &logError ( 20067,
      "<%s> Attribute validation failed. Returning [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }

    &logTrace ( 20068,
    "<%s> Proceeding with second level monitoring", $sMyName );

    #-------------------------------------------------------
    # Upper bound..
    # Time in EPOCH when entry pt started + VCS time limit..
    #-------------------------------------------------------
    $iTillWhen = $^T + MySQL::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         MySQL::SecondLevelMonitor ( $sBaseDir,
                                     $sEnvFile,
                                     $sHostName,
                                     $sMyCnf,
                                     $sMySQLAdmin,
                                     $sMySQLAdminPasswd,
                                     $sMySQLUser,
                                     $iPort,
                                     $sServiceName,
                                     $iTillWhen ) ) ) {

      &logTrace ( 20069,
      "<%s> MySQL instance online with status [%s] after SLM",
      $sMyName, $iReturnCode );
    }
    else {
      &logError ( 20070,
      "<%s> MySQL instance NOT online after SLM. Status [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }
  }

  #------------------------------------------------
  # Exit if external monitor has not been defined..
  #------------------------------------------------
  return $iReturnCode unless defined $sMonitorProgram;

  #---------------------------------------
  # MonitorProgram defined..
  # Separate the executable and the args..
  #---------------------------------------
  my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..

  $sExtCmdArgs = '' unless defined $sExtCmdArgs;   # Prevent undefs..

  if ( $sExtCmd ) {                                # Defined..
    #----------------------------------------
    # Get time bearings, if not got earlier..
    #----------------------------------------
    $iTillWhen ||= $^T + MySQL::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE ==             # and found online..
         ( $iReturnCode =                          # by the status..
         MySQL::ExternalMonitor ( $sMySQLUser,
                                  $sEnvFile,
                                  $sExtCmd,
                                  $sExtCmdArgs,
                                  $iTillWhen ) ) ) {
      &logTrace ( 20071,
      "<%s> MySQL instance found online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $iReturnCode, $sExtCmd, $sExtCmdArgs );
    }
    else {
      &logError ( 20072,
      "<%s> MySQL instance NOT online after executing external monitor
      [%s] with args [%s]. Status [%s]",
      $sMyName, $sExtCmd, $sExtCmdArgs, $iReturnCode );
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int DelayedAttributeCheck ( IN String BaseDir,
                                        IN String EnvFile,
                                        IN String HostName,
                                        IN String MySQLAdmin,
                                        IN String MySQLAdminPasswd,
                                        IN String MySQLUser,
                                        IN Integer Port );

 Description : Just In Time Validation

 Parameters :

 $sBaseDir                  Absolute path to installation directory
 $sEnvFile                  Absolute path to environment file to source
 $sHostName                 Hostname/Bind address of virtual host for MySQL
 $sMySQLAdmin               The dedicated administator for the database server
 $sMySQLAdminPasswd         The encrypted password for the MySQL administrator
 $sMySQLUser                The dedicated OS login used to run the database
 $iPort                     The port on which the database server listens
 
 Returns  : Integer value as defined below:

 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub DelayedAttributeCheck {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sBaseDir,
       $sEnvFile,
       $sHostName,
       $sMySQLAdmin,
       $sMySQLAdminPasswd,
       $sMySQLUser,
       $iPort ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20073, "Subroutine <%s> called with:
  BaseDir                  [%s]
  EnvFile                  [%s]
  HostName                 [%s]
  MySQLAdmin               [%s]
  MySQLAdminPasswd         [%s]
  MySQLUser                [%s]
  Port                     [%s]",

  $sMyName,   $sBaseDir,   $sEnvFile,   $sHostName,   $sMySQLAdmin,
  $sMySQLAdminPasswd, $sMySQLUser, $iPort );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my @lsArgs = qw ( BaseDir EnvFile HostName MySQLAdmin 
                    MySQLAdminPasswd  MySQLUser Port );
  my @lsReqdArgs = qw ( BaseDir HostName MySQLAdmin
                        MySQLAdminPasswd  MySQLUser Port );
  my ( @lsMissing, %hsArgList ) = ();

  @hsArgList { @lsArgs } = @_; # Assign..

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20074,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #--------------------
  # Return, if failed..
  #--------------------
  if ( VCSDef_MonitorExit::ONLINE != $iReturnCode ) {
    return $iReturnCode;
  }

  #------------------------------
  # Validate parameter BaseDir..
  #------------------------------
  if ( ! -d $sBaseDir ) {
    &logError ( 20075,
    "<%s> BaseDir [%s] directory does not exist",
    $sMyName, $sBaseDir );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #------------------------------
  # Validate parameter EnvFile..
  #------------------------------
  if ( $sEnvFile =~ /\S/ && ! -e $sEnvFile ) {
    &logError ( 20076,
    "<%s> EnvFile [%s] file does not exist",
    $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #--------------------------------
  # Validate parameter MySQLUser..
  #--------------------------------
  if ( ! defined getpwnam $sMySQLUser ) {
    &logError ( 20077,
    "<%s> MySQLUser [%s] is not a valid UNIX login",
    $sMyName, $sMySQLUser );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #-------------------------------
  # Validate parameter HostName..
  #-------------------------------
  my ($addr_err, @res) = getaddrinfo($sHostName);
  if ( ! @res ) {
    &logError ( 20078,
    "<%s> HostName [%s] is an invalid Hostname/IPv4/IPv6 address",
    $sMyName, $sHostName );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #---------------------------
  # Validate parameter Port..
  #---------------------------
  if ( ! $iPort ) {
    &logError ( 20079,
    "<%s> Port [%s] has to be greater than 0, if SLM is enabled",
    $sMyName, $iPort );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int SecondLevelMonitor ( IN String BaseDir,
                                     IN String EnvFile,
                                     IN String HostName,
                                     IN String MyCnf,
                                     IN String MySQLAdmin,
                                     IN String MySQLAdminPasswd,
                                     IN String MySQLUser,
                                     IN Integer Port,
                                     IN String ServiceName,
                                     IN Integer TillWhen );

 Description : Does in-depth monitoring, if enabled.

 Parameters :

 $sBaseDir                  Absolute path to installation directory
 $sEnvFile                  Absolute path to environment file to source
 $sHostName                 Hostname/Bind address of virtual host for MySQL
 $sMyCnf                    Absolute path to MySQL configuration file
 $sMySQLAdmin               The dedicated administator for the database server
 $sMySQLAdminPasswd         The encrypted password for the MySQL administrator
 $sMySQLUser                The dedicated OS login used to run the database
 $iPort                     The port on which the database server listens
 $ServiceName              Name of the service to start, stop and check status of MySQL.
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub SecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sBaseDir,
       $sEnvFile,
       $sHostName,
       $sMyCnf,
       $sMySQLAdmin,
       $sMySQLAdminPasswd,
       $sMySQLUser,
       $iPort,
       $sServiceName,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20080, "Subroutine <%s> called with:
  BaseDir                  [%s]
  EnvFile                  [%s]
  HostName                 [%s]
  MyCnf                    [%s]
  MySQLAdmin               [%s]
  MySQLAdminPasswd         [%s]
  MySQLUser                [%s]
  Port                     [%s]
  ServiceName              [%s]
  TillWhen                 [%s]",

  $sMyName,   $sBaseDir,   $sEnvFile,   $sHostName, $sMyCnf,  $sMySQLAdmin,
  $sMySQLAdminPasswd, $sMySQLUser, $iPort, $sServiceName, $iTillWhen );  

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iTimeout, $iExitCode, $sCommand, $sCommandArgs, @lsCmdout ) = ();
  
  #------------------------
  # Return, unless online..
  #------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  #---------------------------------------------
  # Declare monitor command and args variables..
  #---------------------------------------------
  $sCommand = "$sBaseDir/bin/mysqladmin";
  $sCommandArgs = "--defaults-file=$sMyCnf " if ( $sMyCnf =~ /\S/ );
  $sCommandArgs .= "--user=$sMySQLAdmin " .
  		  "--password=$sMySQLAdminPasswd status";

  #-------------------------------
  # Validate and execute command..
  #-------------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20083,
    "<%s> Insufficient time to continue in-depth monitoring.
    Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( SystemD::DetectSystemd()) {
    my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
    if ( $sServiceName !~ /\S/ ) {
      $sServiceName = "mysqld";
      if ( -e "/etc/SuSE-release" ) {
        $sServiceName = "mysql";
      }
    }
    if( $iReturnCode = SystemD::RunSystemdServiceName($sResourceName, $sServiceName, "is-active" )){
      &logTrace ( 20325,
        "<%s> RunSystemd was successful",
        $sMyName );
      $iReturnCode = VCSDef_MonitorExit::ONLINE;
    }
    else {
      $iReturnCode = VCSDef_MonitorExit::OFFLINE;
    }
  }
  elsif ( MySQL::ExecLocalCmd ( $sMySQLUser, 
                                $sEnvFile, 
                                $sCommand, 
                                $sCommandArgs, 
                                \$iExitCode, 
                                $iTimeout, 
                                \@lsCmdout, 
                                $sMySQLAdminPasswd ) ) {
    &logTrace ( 20084,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
    
  }
  elsif ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  else {
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String MySQLUser,
                                  IN String EnvFile,
                                  IN String Cmd,
                                  IN String CmdArgs,
                                  IN Integer TillWhen );

 Description : Handles user defined external monitor program

 Parameters :

 $sMySQLUser                The dedicated OS login used to run the database
 $sEnvFile                  Absolute path to environment file to source
 $sCmd                      User defined utilty to execute
 $sCmdArgs                  Command line arguments
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sMySQLUser,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20085, "Subroutine <%s> called with:
  MySQLUser                [%s]
  EnvFile                  [%s]
  Cmd                      [%s]
  CmdArgs                  [%s]
  TillWhen                 [%s]",

  $sMyName, $sMySQLUser, $sEnvFile, $sCmd, $sCmdArgs, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeout, @lsCmdout ) = ();
  my $iBuf = 5; # Secs..

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE,
                          VCSDef_MonitorExit::TIMEOUT,
                          VCSDef_MonitorExit::TIMEOUT );

  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sCmd ) {
   &logError ( 20086,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
   &logError ( 20087,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $sEnvFile =~ /\S/ && ! -e $sEnvFile ) {
    &logError ( 20088,
    "<%s> EnvFile [%s] does not exist", $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
elsif ( ( $iTimeout = $iTillWhen - ( time + $iBuf ) ) <= 5 ) { # Lack of time..
   &logError ( 20089,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sCmd, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20090,
  "<%s> Executing external monitor with Command [%s] Args [%s] in [%s] secs",
  $sMyName, $sCmd, $sCmdArgs, $iTimeout );

  #----------------------
  # Execute the command..
  #----------------------
  MySQL::ExecLocalCmd ( $sMySQLUser, 
                        $sEnvFile, 
                        $sCmd, 
                        $sCmdArgs, 
                        \$iExitCode, 
                        $iTimeout, 
                        \@lsCmdout );
                               
  #------------------------
  # Report on exit status..
  #------------------------
  if ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iExitCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                             VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( $hiExitCodes { $iExitCode } &&
          VCSDef_MonitorExit::ONLINE ==
          $hiExitCodes { $iExitCode } ) {
    &logTrace ( 20091,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }
  else {
    &logWarn ( 20092,
    "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }

  #--------------------------------
  # Setup appropriate return code..
  #--------------------------------
  $iReturnCode = $hiExitCodes { $iExitCode } ||
                 VCSDef_MonitorExit::UNKNOWN;

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool CreatePidFile ( IN String PidFile,
                                 In Integer Pid );


 Description : Creates PID file with given PID

 Parameters :

 $sPidFile     File to create
 $iPid         PID to store

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CreatePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $iPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20093, "Subroutine <%s> called with:
  PidFile         [%s]
  Pid             [%s]",

  $sMyName, $sPidFile, $iPid );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------
  # Check against undefs..
  #-----------------------
  unless ( defined $iPid && $iPid =~ /^\d+$/o) {
    $iPid ||= 'undef'; # Prevent undefs.
    &logError ( 20094, "<%s> PID [%s] is not a valid positive number",
    $sMyName, $iPid );
    return 0; # Proceed no further..
  }

  #------------------
  # Create PID file..
  #------------------
  &logTrace ( 20095,
  "<%s> Creating PID file [%s] with PID [%s]",
  $sMyName, $sPidFile, $iPid );

  if ( sysopen ( PIDFILE, $sPidFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                 0600 ) ) {
    print PIDFILE $iPid, "\n"; # Update..

    unless ( close ( PIDFILE ) ) {
      &logWarn ( 20096,
      "<%s> Failed to close PidFile [%s] after update: [%s]",
      $sMyName, $sPidFile, $! );
      $iReturnCode = 0;
    }
  }
  else {
    &logWarn ( 20097,
    "<%s> Failed to create PID file [%s] with PID [%s]",
    $sMyName, $sPidFile, $iPid );
    $iReturnCode = 0;
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int FFLMStatus ( IN String PidFile,
                             IN String Proc,
                             IN Integer Uid,
                             OUT RefInt Pid );

 Description: Checks status of Process listed in PidFile, if running

 Parameters :

 $sPidFile       Absolute path to PID file of service
 $sProc          Process string to search for in proc table
 $iUid           UNIX UID of owner
 $riPid          Reference to existing PID

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub FFLMStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $sProc,
       $iUid,
       $riPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20098, "Subroutine <%s> called with:
  PidFile                [%s]
  Proc                   [%s]
  UID                    [%s]
  riPid                  [%s]",

  $sMyName, $sPidFile, $sProc, $iUid, $riPid );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
  $$riPid = 0; # Reset..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iPid, $rhlProc, @lProcessHash ) = ();

  #----------------------------------
  # Start by testing the attributes..
  #----------------------------------
  if ( $sProc !~ /\S/ ) {
    &logError ( 20099, "<%s> Called with empty [%s] process string",
    $sMyName, $sProc );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -e $sPidFile ) {
    &logTrace ( 20100, "<%s> PID file [%s] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -s $sPidFile ) {
    &logWarn ( 20101, "<%s> PID file [%s] is of zero bytes",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #---------------------
  # Return, if offline..
  #---------------------
  return $iReturnCode unless VCSDef_MonitorExit::ONLINE == $iReturnCode;
 
  if ( open ( PIDFILE, $sPidFile ) ) {
    ( $iPid ) = grep { ( s/\n// || 1 ) && # Chomp..
                         s/.*?(\d+)$/$1/  # Get PID..
                } <PIDFILE>;

    $iPid = '' unless defined $iPid;      # Avoid undefs..

    close ( PIDFILE ) ||                  # Close file..
    &logWarn ( 20102,
    "<%s> Could not close PID file [%s] after read [%s]",
    $sMyName, $sPidFile, $! );
  }
  else {
    &logError ( 20103,
    "<%s> Failed to open PID file [%s] for read: [%s]",
    $sMyName, $sPidFile, $! );
    return VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Check if PID is valid..
  #------------------------
  if ( $iPid =~ /\d+/ &&     # If the valid number..
       kill ( 0, $iPid ) ) { # is a PID that's alive..

    &logTrace ( 20104,
    "<%s> PID [%s] from PID file [%s] exists",
    $sMyName, $iPid, $sPidFile );

    #-----------------------
    # Search in proc table..
    #-----------------------
    if ( MySQL::GetPidDetails ( $iPid,
                                 \@lProcessHash ) ) {
      #--------------
      # Got details..
      #--------------
      foreach $rhlProc ( @lProcessHash ) {
        &logTrace ( 20105,
        "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
        $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
        $$rhlProc { COMMAND } );

        #--------------
        # Now compare..
        #--------------
        if ( ( $$rhlProc { UID } == $iUid ) &&       # Owner match..
               $$rhlProc { COMMAND } =~ /$sProc/ ) { # Filter match..
          &logTrace ( 20106,
          "<%s> PID [%s] owned by UID [%s] with ARGS [%s] contains [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $$riPid = $iPid; # Store..
        }
        else {
          &logTrace ( 20107,
          "<%s> Ignoring PID [%s] owned by UID [%s] with ARGS [%s] with string [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
        }
      }
    }
    else {
      &logTrace ( 20108,
      "<%s> Could not get details for PID [%s]",
      $sMyName, $iPid );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  else {
    &logTrace ( 20109, "<%s> PID [%s] does not exist",
    $sMyName, $iPid );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  @lProcessHash = (); # Truncate..
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool KillMySQLProcesses ( IN String BaseDir,
                                      IN String DataDir,
                                      IN String HostName,
                                      IN String MySQLUser, 
                                      IN String ServiceName,
                                      IN String MySQLPidFile );

 Description  : Kills processes for this instance of MySQL
 
 Parameters :

 $sBaseDir                Absolute path to installation directory
 $sDataDir                Absolute path to database directory
 $sHostName               Hostname/Bind address of virtual host for MySQL
 $sMySQLUser              The dedicated OS login used to run the database
 $ServiceName              Name of the service to start, stop and check status of MySQL.
 $MySQLPidFile             Full path of Pid file as specified in service file

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------
=cut

sub KillMySQLProcesses {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sBaseDir,
       $sDataDir,
       $sHostName,
       $sMySQLUser,
       $sServiceName,
       $sMySQLPidFile ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20110, "Subroutine <%s> called with:
  BaseDir                  [%s]
  DataDir                  [%s]
  HostName                 [%s]
  MySQLUser                [%s]
  ServiceName              [%s]
  MySQLPidFile             [%s]",

  $sMyName, $sBaseDir, $sDataDir, $sHostName, $sMySQLUser, $sServiceName, $sMySQLPidFile );
 
  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Set up local variables..
  #-------------------------
  my ( @liGoners, @liUnDead ) = ();

  #---------------------------------------------
  # Get list of processes running for this Uid..
  #---------------------------------------------
  if ( ( caller 1 )[3] eq 'MySQL::Start' ) { # During online..
    @liGoners = @liRunningProcs;              # Get PIDs to kill..
  }
  else {
    MySQL::GetProcState ( \@liGoners,
                           $sBaseDir,
                          $sDataDir,
                          $sHostName,
                          $sMySQLUser,
                          $sServiceName,
                          $sMySQLPidFile,
                          0 ); # Set filtering choice..
  }

  #--------------------------
  # Sort in ascending order..
  #--------------------------
  @liGoners = sort { $a <=> $b } @liGoners;

  #------------------------------------------------------------------
  # We do a two level kill. The SIGINQ ensures  that  the  PID  is  a
  # valid process before we actually issue the SIGKILL. We don't want
  # to report a failure for SIGKILL in case  of  defunct  or  missing
  # processes.
  #------------------------------------------------------------------
  if ( scalar @liGoners ) {
    #----------------------
    # Failed to kill some..
    #----------------------
    if ( @liUnDead = grep {
                         kill ( 0, $_ ) && # Valid..
                       ! kill ( 9, $_ )    # and could NOT be killed..
                     } @liGoners ) {       # From what's found..
      &logError ( 20111,
      "<%s> PIDs not killed [%s] Total not killed [%s]",
      $sMyName, join ( ' ', @liUnDead ), scalar @liUnDead );
      $iReturnCode = 0; # Failure..
    }
    else { # All killed..
      &logInfo ( 20112,
      "<%s> Killed PIDs [%s] Total killed: [%s]",
      $sMyName, join ( ' ', @liGoners ), scalar @liGoners );
    }
  }
  else {
    &logTrace ( 20113,
    "<%s> Could not locate any process this instance of MySQL",
    $sMyName );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ExecLocalCmd ( IN String User,
                                IN String EnvFile,
                                IN String Command,
                                IN String CommandArgs,
                                OUT RefInt ExitCode,
                                IN Integer Timeout,
                                OUT ListReference Cmdout ,
                                IN String HideString );

 Description : Wrapper function for
               Sys::RunWithEnvCmdWithOutputWithTimeOut to collect output
               of the command.
 
 Parameters :

 $sUser            UNIX user to run the command as
 $sEnvFile         Absolute path to location of EnvFile to source
 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $riExitCode       Reference to collect exit code of command
 $iTimeout         Time limit to execute the command
 $rlsCmdout        Stores o/p from command if successful
 $sHideString      String in CommandArgs to hide

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecLocalCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCommand,
       $sCommandArgs,
       $riExitCode,
       $iTimeout,
       $rlsCmdout ,
       $sHideString,
       $sFromDir ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20114, "Subroutine <%s> called with:
  User        [%s]
  EnvFile     [%s]
  Command     [%s]
  CommandArgs [%s]
  riExitCode  [%s]
  Timeout     [%s]
  rlsCmdout   [%s]
  sFromDir    [%s]",

  $sMyName,
  $sUser, $sEnvFile, $sCommand, $sCommandArgs, $riExitCode,
  $iTimeout, $rlsCmdout, $sFromDir );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;

  my ( $iExitCode, $sDispStr, $sCmdDeCr, $sDeCrPasswd ) = ();

  #---------------------
  # Set up for display..
  #---------------------
  $sDispStr = $sCommandArgs;

  #-------------------------
  # Get decrypted password..
  #-------------------------
  if ( defined $sHideString  && $sHideString =~ /\S/ ) {
    #-----------
    # Validate..
    #-----------
    if ( ! -x $sCmdVcsDecrypt ) {
      &logError ( 20115,
      "<%s> Command [%s] is not executable", $sMyName, $sCmdVcsDecrypt );
      return 0; # Failure..
    }

    $sCmdDeCr = $sCmdVcsDecrypt;                          # Get global..
    ( $sDeCrPasswd ) = grep {                             # Get decrypted..
                          s/([^\w-])/\\$1/g || 1          # Escape non-words.. 
                        } readpipe ( "$sCmdDeCr $sHideString" );

    #--------------------
    # Collect exit code..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logError ( 20116, "<%s> Failed to decrypt crypted string [%s]",
      $sMyName, $sHideString );
      return 0; # Failure..
    }

    $sCommandArgs =~ s/$sHideString/$sDeCrPasswd/g;
  }

  #-------------------
  # Set up arguments..
  #-------------------
  my $bSourceUserEnv = 0; # Do not source..
  my $bTimedOut = 0;      # Timeout variable..

  @$rlsCmdout = (); # Reset..

  #-----------
  # Validate..
  #-----------
  if ( ! -e $sCommand ) {
    &logError ( 20117,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCommand ) {
    &logError ( 20118,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iTimeout <= 1 ) {
    &logError ( 20119,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with Args [%s].",
    $sMyName, $iTimeout, $sCommand, $sDispStr );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return if failed.. 

  #----------------------
  # Set LOGNAME to user..
  #----------------------
  my $sOldLogName = $ENV { LOGNAME }; # Backup..
  $ENV { LOGNAME } = $sUser;          # Now set..

  &logTrace ( 20120,
  "<%s> Executing Command [%s] with Args [%s]",
  $sMyName, $sCommand, $sDispStr );

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgs,
                                                           $riExitCode,
                                                           $sUser,
                                                           $rlsCmdout,
                                                           \$bTimedOut,
                                                           $iTimeout,
                                                           undef,
                                                           $sFromDir,
                                                           $sDeCrPasswd );

  #---------------------------
  # Remove trailing newlines..
  #---------------------------
  chomp @$rlsCmdout;

  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) {   # No error with command execution..
    if ( $$riExitCode ) { # Failure..
      &logTrace ( 20121,
      "<%s> Command [%s] with args [%s] returned exit code [%s]",
      $sMyName, $sCommand, $sDispStr, $$riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20122,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sDispStr, $$riExitCode );
    }
  }
  else { # Execution error..
    @$rlsCmdout = (); # Reset..

    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20123,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sDispStr, $iTimeout );
      $$rlsCmdout [0] = 'timeout'; # Timeout..
    }
    else { # Failure due to something else..
      &logError ( 20124,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
    }
  }

  #----------
  # Restore..
  #----------
  defined $sOldLogName ? $ENV { LOGNAME } = $sOldLogName : delete $ENV { LOGNAME };

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetPidDetails ( IN Integer Pid,
                                 OUT RefList ProcList );


 Description : Gets long process details for a known Pid if running

 Parameters :

 $iPid         PID whose details are requested
 $rlProcList   Ref to List to store details
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetPidDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iPid,
       $rlProcList ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20125, "Subroutine <%s> called with:
  Pid             [%s]
  rlProcList      [%s]",

  $sMyName, $iPid, $rlProcList );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------
  # Validate PID..
  #---------------
  unless ( defined $iPid    &&   # Catch undefs..
           $iPid =~ /^\d+$/ &&   # And invalid PIDs..
           kill ( 0, $iPid ) ) { # And defunct ones..
    &logError ( 20126,
    "<%s> PID [%s] is invalid", $sMyName, $iPid );
    return 0;
  }

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sPsCmd, $sPsCmdArgs, $sLang, $sFmt, $sUnix95, $sSwitch, $iExitCode,
       %hsProc ) = ();

  $sSwitch = $^O;    # Get OS..
  @$rlProcList = (); # Truncate..

  #-------------
  # Store LANG..
  #-------------
  $sLang = $ENV { LANG };
  $ENV { LANG } = 'C';

  #---------------------------
  # Switch to the correct OS..
  #---------------------------
  #goto $sSwitch; {
  if ( "solaris" eq $sSwitch ) {
       my $sOSVer = ( uname ) [2]; # Get OS version..
       if ($sOSVer =~ /\d+\.(\d+)/) {
           $sOSVer = $1;
       }
       $sPsCmd = "/usr/bin/ps";
       if ( $sOSVer < 11 ) {
           $sPsCmd = "/usr/ucb/ps" if ( -x "/usr/ucb/ps" );
       }
       $sPsCmdArgs = "axwwl $iPid";
       $sFmt = '[\da-z]+\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';
  }
  elsif ( "hpux" eq $sSwitch ) {
      #-----------------------
      # Force XPG4 behaviour..
      #-----------------------
      $sUnix95 = $ENV { UNIX95 };
      $ENV { UNIX95 } = 'XPG4';

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "-p $iPid -xxlo uid,pid,ppid,args";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';
  }
  elsif ( "aix" eq $sSwitch ) {
      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt =  '[\da-e]+\s+[A-Z]\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';
  }
  elsif ( "linux" eq $sSwitch ) {
      $sPsCmd = "/bin/ps";
      $sPsCmdArgs = "wwo uid,pid,ppid,args $iPid";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';
  }

  &logTrace ( 20127, "<%s> OS [%s] PsCmd [%s] PsCmdArgs [%s]",
  $sMyName, $sSwitch, $sPsCmd, $sPsCmdArgs );

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20128,
    "<%s> ps(1b) command [%s] is not executable", $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  #-------------------------------
  # Fire command and get details..
  #-------------------------------
  @hsProc { qw ( UID PID PPID COMMAND ) } =
  map {
    s/\n//;                             # Chomp..
    s/^\s+//;                           # Strip leading spaces..
    /^$sFmt$/o;                         # Extract fields..
  } readpipe ( "$sPsCmd $sPsCmdArgs" ); # From the command..

  $iExitCode = $? >> 8; # Get exit code..

  #------------------------
  # Restore env variables..
  #------------------------
  defined $sLang ? $ENV { LANG } = $sLang : delete $ENV { LANG };
  defined $sUnix95 ? $ENV { UNIX95 } = $sUnix95 : delete $ENV { UNIX95 };

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logTrace ( 20129,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );

    %hsProc = (); # Truncate..
    return 0;
  }

  #---------------------
  # Fill up the bucket..
  #---------------------
  push ( @$rlProcList, \%hsProc ); # Update..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetSrvrDetails ( IN String BaseDir,
                                  IN String DataDir,
                                  IN String HostName,
                                  IN String ServiceName,
                                  IN String MySQLPidFile,
                                  OUT RefHashList SrvrDetails );

 Description : Gives PID file and process string for given server type

 Parameters  :

 $sBaseDir                  Absolute path to installation directory
 $sDataDir                  Absolute path to database directory
 $sHostName                 Hostname/Bind address of virtual host for MySQL
 $ServiceName              Name of the service to start, stop and check status of MySQL.
 $MySQLPidFile             Full path of Pid file as specified in service file
 $rhlSrvrDetails            Stores details of PID file and process string
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetSrvrDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sBaseDir,
       $sDataDir,
       $sHostName,
       $sServiceName,
       $sMySQLPidFile,
       $rhlSrvrDetails ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20130, "Subroutine <%s> called with:
  BaseDir                  [%s]
  DataDir                  [%s]
  HostName                 [%s]
  ServiceName              [%s]
  MySQLPidFile             [%s]
  SrvrDetails              [%s]",

  $sMyName, $sBaseDir, $sDataDir, $sHostName, $sServiceName, $sMySQLPidFile,  $rhlSrvrDetails );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1; # Set to true..

  #-------------------
  # Define variables..
  #-------------------
  my ( $sKey, $sProc, $sPidFile, $sPIDDir, $sResourceType,
       $sResourceName, $iPat ) = ();

  #-------------------------------------
  # List of search patterns to monitor..
  #-------------------------------------
  $sProc = qr#$sBaseDir.*mysqld(.*--pid-file=$sMySQLPidFile\b|.*--datadir=$sDataDir\b|$)# ;
  if (SystemD::DetectSystemd()) {
    if (  $sServiceName =~ /@(\S+)\b/ ) {
      my $sSuffix = $1;
      $sProc = qr#$sBaseDir.*mysqld.*--defaults-group-suffix=\@$sSuffix\b# ;
    }
  }

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  $sPIDDir = "$VCSLOG/log/tmp/.VRTS$sResourceType/$sResourceName/PID";

  #------------------------------
  # Create PID folder if needed..
  #------------------------------
  if ( -d $sPIDDir ) {
    &logTrace ( 20131, "<%s> PID folder [%s] exists",
    $sMyName, $sPIDDir );
  }
  elsif ( ! mkpath ( $sPIDDir, 0, 0700 ) ) {
    &logWarn ( 20132, "<%s> Failed to create PID folder [%s]",
    $sMyName, $sPIDDir );
    return 0; # Cannot proceed..
  }

  %$rhlSrvrDetails = (); # Truncate..

  #-----------------------------------------------
  # Set appropriate filter strings and PID files..
  #-----------------------------------------------
  $sKey = "mysqld";
  $sPidFile = "$sPIDDir/pidfile.$sKey";
  @{ $$rhlSrvrDetails { $sKey } } = ( $sPidFile, $sProc ); # Store..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetEngineTimeout ( IN Bool UseMargin );

 Description : Calculates engine timeout for entry point
 
 Parameters  :

 $bUseMargin      Flag to indicate if safety margin is requested

                  boolean value in the set of {0,1} where
                  0 No safety margin needed
                  1 Safety margin to be used
 Returns:  Time in secs.
 -------------------------------------------------------------------------

=cut

sub GetEngineTimeout {
  #----------------
  # Get Arguments..
  #----------------
  my ( $bUseMargin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20133, "Subroutine <%s> called with:
  UseMargin  [%s]",
  $sMyName, $bUseMargin );

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iEngineTimeout, $iExitCode ) = ();

  #----------------------------------
  # Grace margins from VCS timeouts..
  #----------------------------------
  my %hiGraceMargin = ( 'MonitorTimeout', 10,   # Seconds..
                        'OnlineTimeout',  10,
                        'OfflineTimeout', 10,
                        'CleanTimeout',   20 );

  #--------------------------------------------------
  # Attempt to get resource level overridden values..
  #--------------------------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;

  #-----------------------------------
  # Query if it has already been set..
  #-----------------------------------
  if ( defined $::giEPTimeout &&    # Defined..
       $::giEPTimeout =~ /^\d+/ ) { # to an integer..
    $iEngineTimeout = $::giEPTimeout;

    &logTrace ( 20134,
    "<%s> Using AEPTimeout [%s] as EP timeout value",
    $sMyName, $iEngineTimeout );
  }
  else {
    #-----------------------------------------
    # Form query string based on entry point..
    #-----------------------------------------
    $sCommand = $sCmdHares; # Get global..
    $sCommandArgs = "-value $sResourceName $sAttribute";
  
    #----------------------------------
    # Fire command and collect output..
    #----------------------------------
    ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                      =~ s/\n//;
  
    #--------------------
    # Check exit status..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logTrace ( 20135,
      "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    }
    elsif ( ! $iEngineTimeout ) { # Undefined value..
      &logTrace ( 20136, "<%s> Could not get [%s] from engine",
      $sMyName, $sAttribute );
    }
    else { # Success..
      &logTrace ( 20137, "<%s> Got [%s] as [%s] sec from engine",
      $sMyName, $sAttribute, $iEngineTimeout );
    }
  
    #-----------------------------------------
    # Form query string based on entry point..
    #-----------------------------------------
    unless ( $iEngineTimeout ) {
      $sCommand = $sCmdHatype; # Get global..
      ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
      $sResourceType = "MySQL" if $sResourceType !~ /\w/; # Ensure..
      $sCommandArgs = "-value $sResourceType $sAttribute";
    
      #----------------------------------
      # Fire command and collect output..
      #----------------------------------
      ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs" ) ) =~ s/\n//;
    
      #--------------------
      # Check exit status..
      #--------------------
      if ( $iExitCode = $? >> 8 ) {
        &logTrace ( 20138,
        "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
        $sMyName, $sCommand, $sCommandArgs, $iExitCode );
        $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
      }
      elsif ( ! $iEngineTimeout ) { # Empty value..
        &logTrace ( 20139, "<%s> Could not get [%s] from engine",
        $sMyName, $sAttribute );
        $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
      }
      else { # Success..
        &logTrace ( 20140, "<%s> Got [%s] as [%s] sec from engine",
        $sMyName, $sAttribute, $iEngineTimeout );
      }
    }
  }

  #----------------------------
  # Apply margin if requested.. 
  #----------------------------
  if ( $bUseMargin     &&                                    # Margin requested..
       $iEngineTimeout >= $hiGraceMargin { $sAttribute } ) { # And valid..
    $iEngineTimeout -= $hiGraceMargin { $sAttribute };       # Apply..

    &logTrace ( 20141,
    "<%s> Setting up [%s] sec margin. Returning [%s]",
    $sMyName, $hiGraceMargin { $sAttribute }, $iEngineTimeout );
  }

  return ( $iEngineTimeout ); # Secs..
}

1; # So PERL knows its included..
